#include <iostream>
#include <vector>
#include <fstream>
#include <set>
#include <cmath>
#include <ctime>
#include <algorithm>
#include <cstring>


using namespace std; 

// Some definitions
#define SINGLE_PRECISION

#ifdef SINGLE_PRECISION
    #define REAL    float
#else
    #define REAL    double
#endif

#include "predicates.h"

typedef enum {
    BINARY, 
    TEXT
} Mode; 

// Some constants
const char* vertex_fn       = "data.vertex"; 
const char* constraint_fn   = "data.constraint"; 

int noVertices        = 10000000; 
int noConstraints     = 1000000; 
int seed              = 1; 

const Mode outputMode       = BINARY; 

const int gridSize          = 256; 
#define DISTRIBUTION 0 // 0-uniform, 1-Gaussian

// Random generator
unsigned long z, w, jsr, jcong; // Seeds
void randinit(unsigned long x_) 
{ z =x_; w = x_; jsr = x_; jcong = x_; }
unsigned long znew() 
{ return (z = 36969 * (z & 0xfffful) + (z >> 16)); }
unsigned long wnew() 
{ return (w = 18000 * (w & 0xfffful) + (w >> 16)); }
unsigned long MWC()  
{ return ((znew() << 16) + wnew()); }
unsigned long SHR3()
{ jsr ^= (jsr << 17); jsr ^= (jsr >> 13); return (jsr ^= (jsr << 5)); }
unsigned long CONG() 
{ return (jcong = 69069 * jcong + 1234567); }
unsigned long rand_int()         // [0,2^32-1]
{ return ((MWC() ^ CONG()) + SHR3()); }

REAL random()     // [0,1]
{ return ((REAL) rand_int() / (REAL(ULONG_MAX)+1)); }

// Helper functions
void GaussianRand(REAL *x, REAL *y)
{
    REAL x1, x2, w;

    do {
        x1 = 2.0 * random() - 1.0;
        x2 = 2.0 * random() - 1.0;
        w = x1 * x1 + x2 * x2;
    } while ( w >= 1.0 );

    w = sqrt( (-2.0 * log( w ) ) / w );
    *x = x1 * w;
    *y = x2 * w;
}

bool inRange( REAL r0, REAL r1, REAL value )
{
    return ( ( value >= r0 && value <= r1 ) || ( value >= r1 && value <= r0 ) ); 
}

void parseArguments( int argc, char *argv[] )
{
    for ( int i = 0; i < argc; i++ )
    {
        if ( strcmp( argv[i], "-n" ) == 0 ) 
        {
            noVertices = atoi( argv[i+1] ); 
            i++; 

        } else if ( strcmp( argv[i], "-c" ) == 0 )
        {
            noConstraints = atoi( argv[i+1] ); 
            i++; 

        } else if ( strcmp( argv[i], "-s" ) == 0 )
        {
            seed = atoi( argv[i+1] ); 
            i++; 
        }
    }
}

// Helper classes
class Constraint
{
public: 
    int a, b; 

    Constraint( int src, int dest ) : a(src), b(dest) {}; 
};

typedef vector< Point2D >       Point2DVec; 
typedef vector< Constraint >    ConstraintVec; 
typedef vector< int >           IntVec; 

class Grid
{
public: 
    vector< vector< IntVec > >  pointMap; 
    vector< vector< IntVec > >  constraintMap; 

    vector< vector< bool > > tempMap; 

    Point2DVec           pointVec; 
    ConstraintVec        constraintVec; 

    const int size; 

    class Cell {
    public: 
        int x, y; 

        Cell( const int px, const int py ) : x( px ), y( py ) {}
        Cell( Point2D p ) : x( int(p.x) ), y( int(p.y) ) {}

        bool operator== ( const Cell& c ) const
        {
            return ( x == c.x && y == c.y ); 
        }

        bool isIntersecting( const Point2D& pa, const Point2D& pb ) const
        {
            if ( *this == pa || *this == pb )
                return true; 

            // The cell must intersect the bounding box of the segment
            if ( !inRange( pa.x, pb.x, x ) && !inRange( pa.x, pb.x, x + 1 ) &&
                 !inRange( x, x + 1, pa.x ) && !inRange( x, x + 1, pb.x ) )
                return false; 

            if ( !inRange( pa.y, pb.y, y ) && !inRange( pa.y, pb.y, y + 1 ) &&
                 !inRange( y, y + 1, pa.y ) && !inRange( y, y + 1, pb.y ) )
                return false; 

            // One of the diagonal of the cell must intersects
            REAL ccw0 = counterclockwise_inexact( pa, pb, Point2D(x, y) ); 
            REAL ccw2 = counterclockwise_inexact( pa, pb, Point2D(x + 1, y + 1) ); 

            if ( ccw0 * ccw2 <= 0 )
                return true; 

            REAL ccw1 = counterclockwise_inexact( pa, pb, Point2D(x + 1, y) ); 
            REAL ccw3 = counterclockwise_inexact( pa, pb, Point2D(x, y + 1) ); 

            if ( ccw1 * ccw3 <= 0 )
                return true; 

            return false; 
        }
    };

    typedef vector< Cell >      CellVec; 

    Grid(const int width) : size(width)
    {
        constraintMap   = vector< vector<IntVec> >( width, vector<IntVec>( width ) ); 
        pointMap        = vector< vector<IntVec> >( width, vector<IntVec>( width ) ); 
        tempMap         = vector< vector<bool> >( width, vector<bool>( width, false ) ); 
    }

    void addPointToMap( int i ) 
    {
        const Point2D& p = pointVec[i]; 
        const Cell c( p ); 

        pointMap[c.x][c.y].push_back( i ); 
    }

    void addPoint( const Point2D p ) 
    {
        pointVec.push_back( p ); 

        addPointToMap( pointVec.size() - 1 ); 
    }

    REAL nextReal() const
    {     

		REAL value = random() * size; 

        if ( value >= size ) 
            value -= 1.0; 

        return value; 
    } 

	Point2D nextReal2() const
	{
		REAL x, y;
		
		do
        {
            GaussianRand( &x, &y );
        } while(x < -3 || x > 3 || y < -3 || y > 3);

        x = ( (x + 3.0) / 6.0 ) * size;
        y = ( (y + 3.0) / 6.0 ) * size;

        if ( x >= size ) 
            x -= 1.0; 

        if ( y >= size ) 
            y -= 1.0; 

		return Point2D( x, y );
	}

    bool pointInMap( const Point2D& p ) const
    {
        int cx = int( p.x ); 
        int cy = int( p.y ); 

        const IntVec& list = pointMap[cx][cy]; 

        for ( unsigned int i = 0; i < list.size(); i++ ) 
            if ( p == pointVec[ list[i] ] )
                return true;

        return false; 
    }

    bool pointNearConstraint( const Point2D& p, const int c ) const
    {
        const Constraint& constraint    = constraintVec[c]; 
        const Point2D& p0               = pointVec[ constraint.a ]; 
        const Point2D& p1               = pointVec[ constraint.b ]; 

        if ( !inRange( p0.x, p1.x, p.x ) ||
             !inRange( p0.y, p1.y, p.y ) )
             return false; 

        return ( fabs( counterclockwise_inexact( p0, p1, p ) ) < 0.1 ); 
    }

    // A conservative check to see if the point is near any constraints
    bool pointNearConstraints( const Point2D& p ) const
    {
        int cx = int( p.x ); 
        int cy = int( p.y ); 

        const IntVec& list = constraintMap[cx][cy]; 

        for ( unsigned int i = 0; i < list.size(); i++ ) 
            if ( pointNearConstraint( p, list[i] ) )
                return true;

        // FULL CHECK, DEBUG
        //for ( unsigned int i = 0; i < constraintVec.size(); i++ )         
        //{
        //    const Constraint& c = constraintVec[i]; 
        //    const Point2D& pa   = pointVec[ c.a ]; 
        //    const Point2D& pb   = pointVec[ c.b ]; 

        //    if ( p.x < pa.x && p.x < pb.x )
        //        continue; 

        //    if ( p.x > pa.x && p.x > pb.x )
        //        continue; 

        //    if ( p.y < pa.y && p.y < pb.y )
        //        continue; 

        //    if ( p.y > pa.y && p.y > pb.y )
        //        continue; 

        //    if ( counterclockwise( pa, pb, p ) == 0 )
        //        printf( " Point %i on constraint %i\n", pointVec.size() - 1, i ); 
        //}
        ///

        return false; 
    }

    // Generate a point. The point is NOT added to the data-structure
    Point2D genPoint() const
    {
        Point2D p( 0.0, 0.0 ); 

        do 
        {
            if(DISTRIBUTION == 0)
				p = Point2D( nextReal(), nextReal() ); 
			if(DISTRIBUTION == 1)
				p = nextReal2();

        } while ( pointInMap( p ) || pointNearConstraints( p ) ); 

        return p; 
    }

    CellVec cellsIntersectConstraint( const Point2D& a, const Point2D& b ) 
    {
        const int noNeighbors = 4; 

        const int goX[] = { -1,  0, 0, 1 }; 
        const int goY[] = {  0, -1, 1, 0 }; 

        CellVec queue; 
        unsigned int first = 0; 

        Cell start( a ); 
        queue.push_back( start ); 

        tempMap[ start.x ][ start.y ] = true; 

        while ( first < queue.size() ) 
        {
            Cell c = queue[ first ]; 
            first++; 

            for ( int i = 0; i < noNeighbors; i++ ) 
            {
                Cell n( c.x + goX[i], c.y + goY[i] ); 

                if ( n.x >= 0 && n.y >= 0 && n.x < size && n.y < size &&
                     tempMap[ n.x ][ n.y ] == false && 
                     n.isIntersecting( a, b ) 
                   )
                {
                    queue.push_back( n ); 
                    tempMap[ n.x ][ n.y ] = true; 
                }
            }
        }

        // Clean the tempMap
        for ( unsigned int i = 0; i < queue.size(); i++ )
            tempMap[ queue[i].x ][ queue[i].y ] = false; 

        return queue; 
    }

    Point2D midpoint( const Point2D& pa, const Point2D& pb ) const
    {
        return Point2D( ( pa.x + pb.x ) / 2, ( pa.y + pb.y ) / 2 ); 
    }

    bool isIntersecting( const Point2D& p0, const Point2D& p1, const Point2D& q0, const Point2D& q1 )
    {
        REAL ccw0 = counterclockwise( p0, p1, q0 ); 
        REAL ccw1 = counterclockwise( p0, p1, q1 ); 

        if ( ccw0 * ccw1 > 0.0 )
            return false; 

        REAL ccw2 = counterclockwise( q0, q1, p0 ); 
        REAL ccw3 = counterclockwise( q0, q1, p1 ); 

        if ( ccw2 * ccw3 > 0.0 )
            return false; 

        return true; 
    }

    // Generate a constraint. The constraint is added to the data-structure
    void genConstraint()
    {
        // Gen the two vertices
        Point2D pa = genPoint();
        Point2D pb(0, 0); 
        
        do {
            pb = genPoint(); 
        } while ( pa == pb ); 

        // Traverse from pa to pb using a flooding style for simplicity
        CellVec crossedCells = cellsIntersectConstraint( pa, pb ); 

        int i = (int) crossedCells.size() - 1; 

        while (i >= 0 )
        {
            const Cell& c               = crossedCells[i]; 
            const IntVec& constraints   = constraintMap[ c.x ][ c.y ]; 

            // Go through all the constraints crossing this cell
            for ( unsigned int j = 0; j < constraints.size(); j++ )
            {
                const Constraint& nc    = constraintVec[ constraints[j] ]; 
                const Point2D& pc       = pointVec[ nc.a ]; 
                const Point2D& pd       = pointVec[ nc.b ]; 

                // Repeatly half the constraint if there's an intersection
                bool intersected = false; 

                while ( isIntersecting( pa, pb, pc, pd ) )
                {
                    intersected = true; 
                    const Point2D mid = midpoint( pa, pb ); 

                    if ( pa == mid || pb == mid ) 
                    {
                        pa = genPoint(); 
                        pb = genPoint(); 
                        break; 
                    }
                    else
                        pb = mid; 
                }

                if ( intersected ) 
                {
                    // Refresh the crossedCells list
                    crossedCells = cellsIntersectConstraint( pa, pb ); 
                    i = crossedCells.size(); 
                    break; 
                }
            }

            i--; 
        }

        // VERIFICATION (DEBUG)
        //for ( unsigned int i = 0; i < constraintVec.size(); i++ )
        //{
        //    const Constraint& nc    = constraintVec[ i ]; 
        //    const Point2D& pc       = pointVec[ nc.a ]; 
        //    const Point2D& pd       = pointVec[ nc.b ]; 

        //    if ( isIntersecting( pa, pb, pc, pd ) )
        //    {
        //        printf( " Error: %i intersecting %i\n", constraintVec.size(), i ); 
        //    }
        //}
        /////////

        // OK, add the constraint to the data-structure
        Constraint c( pointVec.size(), pointVec.size() + 1 ); 
        addPoint( pa ); 
        addPoint( pb ); 

        crossedCells = cellsIntersectConstraint( pa, pb ); 

        for ( unsigned int i = 0; i < crossedCells.size(); i++ )
            constraintMap[ crossedCells[i].x ][ crossedCells[i].y ].push_back( constraintVec.size() ); 

        constraintVec.push_back( c ); 
    }
};

void outputVerticesText( const char* filename, const Point2DVec& points )
{
    cout << "Writing vertices file..." << endl; 

    ofstream vout( filename, ios_base::out ); 

    vout << points.size() << endl; 

    vout.precision( 10 ); 

    for ( unsigned int i = 0; i < points.size(); i++ )
        vout << i << " " << points[i].x << " " << points[i].y << endl; 

    vout.close(); 
}

void outputConstraintsText( const char* filename, const ConstraintVec& constraints )
{
    cout << "Writing constraints file..." << endl; 

    ofstream sout( filename, ios_base::out ); 

    sout << constraints.size(); 

    for ( unsigned int i = 0; i < constraints.size(); i++ )
        sout << i << " " << constraints[i].a << " " << constraints[i].b << endl; 

    sout.close(); 
}

void outputVerticesBinary( const char* filename, const Point2DVec& points )
{
    cout << "Writing vertices file..." << endl; 

    ofstream vout( filename, ios_base::out | ios_base::binary ); 

    const unsigned int noPoints = points.size(); 

    vout.write( (const char*) &noPoints, sizeof( unsigned int ) ); 
    vout.write( (const char*) &points[0], sizeof( Point2D ) * noPoints ); 

    vout.close(); 
}

void outputConstraintsBinary( const char* filename, const ConstraintVec& constraints )
{
    cout << "Writing constraints file..." << endl; 

    ofstream sout( filename, ios_base::out | ios_base::binary ); 

    const unsigned int noConstraints = constraints.size(); 

    sout.write( (const char*) &noConstraints, sizeof( unsigned int ) ); 
    sout.write( (const char*) &constraints[0], sizeof( Constraint ) * noConstraints ); 

    sout.close(); 
}

// MAIN program
void main( int argc, char *argv[] ) 
{
    parseArguments( argc, argv ); 

    cout << "  noVertices    = " << noVertices << endl; 
    cout << "  noConstraints = " << noConstraints << endl; 
    cout << "  seed          = " << seed << endl; 

    gpudt_exactinit(); 

    if ( noConstraints * 2 > noVertices ) 
    {
        cout << "Invalid number of constraints" << endl; 
        exit(0); 
    }    

    randinit(seed); 

    // Generate test data
    Grid grid( gridSize ); 

    clock_t startTime = clock(); 

    cout << "Generating constraints..." << endl; 

    for ( int i = 0; i < noConstraints; i++ ) 
    {
        if ( clock() - startTime > CLOCKS_PER_SEC ) 
        {
            cout << ( ( (long long) i ) * 1000 / noConstraints) / 10.0 << endl; 
            startTime = clock(); 
        }

        grid.genConstraint(); 
    }

    cout << "Done!" << endl; 
    cout << endl; 

    cout << "Generating points..." << endl; 
    int moreVertices = noVertices - noConstraints * 2; 

    for ( int i = 0; i < moreVertices; i++ )
    {
        if ( clock() - startTime > CLOCKS_PER_SEC ) 
        {
            cout << ( ( (long long) i ) * 1000 / moreVertices) / 10.0 << endl; 
            startTime = clock(); 
        }

        Point2D p = grid.genPoint(); 

        grid.addPoint( p ); 
    }

    cout << "Done!" << endl; 

    // Write to file
    if ( outputMode == TEXT ) 
    {
        outputVerticesText( vertex_fn, grid.pointVec ); 
        outputConstraintsText( constraint_fn, grid.constraintVec); 
    }
    else
    {
       
		char s[100];
		itoa(seed,s,10);
		string st1,st;
		st = "data";
		st += s;
		st +=".vertex";
		strcpy(s, st.c_str() );
		outputVerticesBinary( s, grid.pointVec ); 

        itoa(seed,s,10); 
        st = "data";
		st += s;
		st +=".constraint";
		strcpy(s, st.c_str() );

        outputConstraintsBinary( s, grid.constraintVec); 
    }
}